#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.symbol_util as symbol_util
import octobot_commons.enums as commons_enums
import octobot_trading.enums as trading_enums
import octobot_trading.api as trading_api


def set_plot_orders(ctx, value):
    ctx.plot_orders = value


async def store_orders(ctx, orders,
                       chart=commons_enums.PlotCharts.MAIN_CHART.value,
                       x_multiplier=1000,
                       mode="markers",
                       kind="scattergl"):
    order_data = [
        {
            "x": order.creation_time * x_multiplier,
            "text": f"{order.order_type.name} {order.origin_quantity} {order.currency} at {order.origin_price}",
            "id": order.order_id,
            "symbol": order.symbol,
            "type": order.order_type.name if order.order_type is not None else 'Unknown',
            "volume": float(order.origin_quantity),
            "y": float(order.created_last_price),
            "cost": float(order.total_cost),
            "state": order.state.state.value if order.state is not None else 'Unknown',
            "chart": chart,
            "kind": kind,
            "side": order.side.value,
            "mode": mode,
            "color": "red" if order.side is trading_enums.TradeOrderSide.SELL else "blue",
            "size": "10",
            "shape": "arrow-bar-left" if order.side is trading_enums.TradeOrderSide.SELL else "arrow-bar-right"
        }
        for order in orders
    ]
    await ctx.orders_writer.log_many(commons_enums.DBTables.ORDERS.value, order_data)


async def store_trade(ctx,
                      trade_dict,
                      chart=commons_enums.PlotCharts.MAIN_CHART.value,
                      x_multiplier=1000,
                      kind="scattergl",
                      mode="markers",
                      exchange_manager=None,
                      writer=None):
    exchange_manager = exchange_manager or ctx.exchange_manager
    tag = f"{trade_dict[trading_enums.ExchangeConstantsOrderColumns.TAG.value]} " \
        if trade_dict[trading_enums.ExchangeConstantsOrderColumns.TAG.value] else ""
    symbol = trade_dict[trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value]
    trade_side = trade_dict[trading_enums.ExchangeConstantsOrderColumns.SIDE.value]
    is_using_positions = False
    color = shape = None
    if exchange_manager.is_future:
        positions = exchange_manager.exchange_personal_data.positions_manager.get_symbol_positions(symbol=symbol)
        if positions:
            is_using_positions = True
            # trading_side = next(iter(positions)).side
            # if trading_side is trading_enums.PositionSide.LONG:
            if "stop_loss" in trade_dict[trading_enums.ExchangeConstantsOrderColumns.TYPE.value]:
                shape = "x"
                color = "orange"
            elif trade_dict[trading_enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value] is True:
                if trade_side == trading_enums.TradeOrderSide.SELL.value:
                    # long tp
                    color = "magenta"
                    shape = "arrow-bar-left"
                else:
                    # short tp
                    color = "blue"
                    shape = "arrow-bar-left"
            else:
                if trade_side == trading_enums.TradeOrderSide.BUY.value:
                    # long entry
                    color = "green"
                    shape = "arrow-bar-right"
                else:
                    # short entry
                    color = "red"
                    shape = "arrow-bar-right"

    if not is_using_positions:
        if trade_side == trading_enums.TradeOrderSide.BUY.value:
            color = "blue"
            shape = "arrow-bar-right"
        elif "stop_loss" in trade_dict[trading_enums.ExchangeConstantsOrderColumns.TYPE.value]:
            color = "orange"
            shape = "x"
        else:
            color = "magenta"
            shape = "arrow-bar-left"

    trade_data = {
        "x": trade_dict[trading_enums.ExchangeConstantsOrderColumns.TIMESTAMP.value] * x_multiplier,
        "text": f"{tag}{trade_dict[trading_enums.ExchangeConstantsOrderColumns.TYPE.value]} "
                f"{trade_dict[trading_enums.ExchangeConstantsOrderColumns.SIDE.value]} "
                f"{trade_dict[trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value]} "
                f"{symbol_util.split_symbol(trade_dict[trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value])[0]} "
                f"at {trade_dict[trading_enums.ExchangeConstantsOrderColumns.PRICE.value]}",
        "id": trade_dict[trading_enums.ExchangeConstantsOrderColumns.ID.value],
        "symbol": trade_dict[trading_enums.ExchangeConstantsOrderColumns.SYMBOL.value],
        "type": trade_dict[trading_enums.ExchangeConstantsOrderColumns.TYPE.value],
        "volume": float(trade_dict[trading_enums.ExchangeConstantsOrderColumns.AMOUNT.value]),
        "y": float(trade_dict[trading_enums.ExchangeConstantsOrderColumns.PRICE.value]),
        "cost": float(trade_dict[trading_enums.ExchangeConstantsOrderColumns.COST.value]),
        "state": trade_dict[trading_enums.ExchangeConstantsOrderColumns.STATUS.value],
        "chart": chart,
        "kind": kind,
        "side": trade_dict[trading_enums.ExchangeConstantsOrderColumns.SIDE.value],
        "mode": mode,
        "shape": shape,
        "color": color,
        "size": "10",
        "fees_amount": float(trade_dict[trading_enums.ExchangeConstantsOrderColumns.FEE.value]
                             [trading_enums.ExchangeConstantsFeesColumns.COST.value] if
                             trade_dict[trading_enums.ExchangeConstantsOrderColumns.FEE.value] else 0),
        "fees_currency": trade_dict[trading_enums.ExchangeConstantsOrderColumns.FEE.value][
            trading_enums.ExchangeConstantsFeesColumns.CURRENCY.value]
        if trade_dict[trading_enums.ExchangeConstantsOrderColumns.FEE.value] else "",
    }
    await writer.log(commons_enums.DBTables.TRADES.value, trade_data)


async def store_transactions(ctx,
                             transactions,
                             chart=commons_enums.PlotCharts.MAIN_CHART.value,
                             x_multiplier=1000,
                             kind="scattergl",
                             mode="markers",
                             y_data=None,
                             writer=None):
    y_data = y_data or [0] * len(transactions)
    transactions_data = [
        {
            "x": transaction.creation_time * x_multiplier,
            "type": transaction.transaction_type.value,
            "id": transaction.transaction_id,
            "symbol": transaction.symbol,
            "currency": transaction.currency,
            "quantity": float(transaction.quantity) if hasattr(transaction, "quantity") else None,
            "order_id": transaction.order_id if hasattr(transaction, "order_id") else None,
            "funding_rate": float(transaction.funding_rate) if hasattr(transaction, "funding_rate") else None,
            "realised_pnl": float(transaction.realised_pnl) if hasattr(transaction, "realised_pnl") else None,
            "transaction_fee": float(transaction.transaction_fee) if hasattr(transaction, "transaction_fee") else None,
            "closed_quantity": float(transaction.closed_quantity) if hasattr(transaction, "closed_quantity") else None,
            "cumulated_closed_quantity": float(transaction.cumulated_closed_quantity)
            if hasattr(transaction, "cumulated_closed_quantity") else None,
            "first_entry_time": float(transaction.first_entry_time) * x_multiplier
            if hasattr(transaction, "first_entry_time") else None,
            "average_entry_price": float(transaction.average_entry_price)
            if hasattr(transaction, "average_entry_price") else None,
            "average_exit_price": float(transaction.average_exit_price)
            if hasattr(transaction, "average_exit_price") else None,
            "order_exit_price": float(transaction.order_exit_price)
            if hasattr(transaction, "order_exit_price") else None,
            "leverage": float(transaction.leverage) if hasattr(transaction, "leverage") else None,
            "trigger_source": transaction.trigger_source.value if hasattr(transaction, "trigger_source") else None,
            "side": transaction.side.value if hasattr(transaction, "side") else None,
            "y": y_data[index],
            "chart": chart,
            "kind": kind,
            "mode": mode
        }
        for index, transaction in enumerate(transactions)
    ]
    writer = writer or ctx.transactions_writer
    await writer.log_many(commons_enums.DBTables.TRANSACTIONS.value, transactions_data)


async def save_metadata(writer, metadata):
    await writer.log(
        commons_enums.DBTables.METADATA.value,
        metadata
    )


async def save_portfolio(writer, context):
    await writer.log(
        commons_enums.DBTables.PORTFOLIO.value,
        trading_api.get_portfolio(context.exchange_manager, as_decimal=False)
    )


async def clear_run_data(writer):
    await _clear_table(writer, commons_enums.DBTables.METADATA.value, flush=False)
    await _clear_table(writer, commons_enums.DBTables.PORTFOLIO.value, flush=False)
    await writer.clear()


async def clear_orders_cache(writer):
    await _clear_table(writer, commons_enums.DBTables.ORDERS.value)
    await writer.clear()


async def clear_trades_cache(writer):
    await _clear_table(writer, commons_enums.DBTables.TRADES.value)
    await writer.clear()


async def clear_transactions_cache(writer):
    await _clear_table(writer, commons_enums.DBTables.TRANSACTIONS.value)
    await writer.clear()


async def clear_all_tables(writer_reader):
    for table in await writer_reader.tables():
        await _clear_table(writer_reader, table)
    await writer_reader.clear()


async def _clear_table(writer, table, flush=True):
    await writer.delete(table, None)
    if flush:
        await writer.flush()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from copy import deepcopy

import pytest

from octobot_trading.exchange_data.order_book.order_book_manager import OrderBookManager
from octobot_trading.enums import ExchangeConstantsOrderBookInfoColumns as ECOBIC
from octobot_trading.enums import TradeOrderSide
from tests.test_utils.random_numbers import random_price_list, random_price, random_quantity, random_order_book_side
from tests.test_utils.random_numbers import random_timestamp
from tests import event_loop

pytestmark = pytest.mark.asyncio


@pytest.fixture()
async def order_book_manager():
    ob_manager = OrderBookManager()
    await ob_manager.initialize()
    return ob_manager


async def test_init(order_book_manager):
    assert not order_book_manager.order_book_initialized
    assert order_book_manager.ask_quantity == 0
    assert order_book_manager.ask_price == 0
    assert order_book_manager.bid_quantity == 0
    assert order_book_manager.bid_price == 0


async def test_reset(order_book_manager):
    order_book_manager.bids = random_price_list(10)
    order_book_manager.bid_price = random_price()
    order_book_manager.bid_quantity = random_price()
    order_book_manager.reset_order_book()
    assert order_book_manager.bids == []
    assert order_book_manager.bid_quantity == 0
    assert order_book_manager.bid_price == 0


async def test_handle_new_books(order_book_manager):
    ts = random_timestamp()
    asks = random_order_book_side(count=100)
    bids = random_order_book_side(count=100)
    order_book_manager.handle_new_books(asks, bids, timestamp=ts)
    assert order_book_manager.order_book_initialized
    assert order_book_manager.timestamp == ts
    assert order_book_manager.get_ask()
    assert order_book_manager.get_bid()


async def test_order_book_ticker_update(order_book_manager):
    b_price = random_price()
    a_price = random_price()
    b_quantity = random_quantity()
    a_quantity = random_quantity()
    order_book_manager.order_book_ticker_update(a_quantity, a_price, b_quantity, b_price)
    assert order_book_manager.ask_quantity == a_quantity
    assert order_book_manager.ask_price == a_price
    assert order_book_manager.bid_quantity == b_quantity
    assert order_book_manager.bid_price == b_price


async def test_handle_new_book(order_book_manager):
    """
    Handle ccxt request
    {
    'bids': [
        [ price, amount ], // [ float, float ]
        [ price, amount ],
        ...
    ],
    'asks': [
        [ price, amount ],
        [ price, amount ],
        ...
    ],
    'timestamp': 1499280391811, // Unix Timestamp in milliseconds (seconds * 1000)
    'datetime': '2017-07-05T18:47:14.692Z', // ISO8601 datetime string with milliseconds
    'nonce': 1499280391811, // an increasing unique identifier of the orderbook snapshot
    }
    """
    ts = random_timestamp()
    asks = random_order_book_side(count=100)
    bids = random_order_book_side(count=100)
    order_book_manager.handle_new_book({
        ECOBIC.ASKS.value: asks,
        ECOBIC.BIDS.value: bids,
        ECOBIC.TIMESTAMP.value: ts
    })
    assert order_book_manager.timestamp == ts
    assert order_book_manager.get_ask()
    assert order_book_manager.get_bid()


async def test_handle_book_adds(order_book_manager):
    order_book_manager.handle_book_adds([
        get_test_order(TradeOrderSide.BUY.value, "1"),
        get_test_order(TradeOrderSide.BUY.value, "2"),
        get_test_order(TradeOrderSide.SELL.value, "3")
    ])
    assert get_order_at_id_in_order_list("1", order_book_manager.bids)
    assert get_order_at_id_in_order_list("2", order_book_manager.bids)
    assert not get_order_at_id_in_order_list("1", order_book_manager.asks)
    assert not get_order_at_id_in_order_list("2", order_book_manager.asks)
    assert not get_order_at_id_in_order_list("3", order_book_manager.bids)
    assert get_order_at_id_in_order_list("3", order_book_manager.asks)
    assert not get_order_at_id_in_order_list("4", order_book_manager.asks)
    order_book_manager.reset_order_book()
    assert not get_order_at_id_in_order_list("1", order_book_manager.bids)
    assert not get_order_at_id_in_order_list("2", order_book_manager.bids)
    assert not get_order_at_id_in_order_list("3", order_book_manager.asks)


async def test_handle_book_deletes(order_book_manager):
    order_2 = get_test_order(TradeOrderSide.BUY.value, "2")
    order_3 = get_test_order(TradeOrderSide.SELL.value, "3")
    order_5 = get_test_order(TradeOrderSide.SELL.value, "5")
    order_book_manager.handle_book_adds([
        get_test_order(TradeOrderSide.BUY.value, "1"),
        order_2,
        order_3,
        get_test_order(TradeOrderSide.SELL.value, "4"),
        order_5
    ])
    assert get_order_at_id_in_order_list("1", order_book_manager.bids)
    assert get_order_at_id_in_order_list("2", order_book_manager.bids)
    assert get_order_at_id_in_order_list("3", order_book_manager.asks)
    assert get_order_at_id_in_order_list("4", order_book_manager.asks)
    assert get_order_at_id_in_order_list("5", order_book_manager.asks)
    order_book_manager.handle_book_deletes([order_5])
    assert get_order_at_id_in_order_list("4", order_book_manager.asks)
    assert not get_order_at_id_in_order_list("5", order_book_manager.asks)
    order_book_manager.handle_book_deletes([order_2, order_3])
    assert get_order_at_id_in_order_list("1", order_book_manager.bids)
    assert not get_order_at_id_in_order_list("2", order_book_manager.bids)
    assert not get_order_at_id_in_order_list("3", order_book_manager.asks)
    assert get_order_at_id_in_order_list("4", order_book_manager.asks)
    assert not get_order_at_id_in_order_list("5", order_book_manager.asks)


async def test_handle_book_updates(order_book_manager):
    order_3 = get_test_order(TradeOrderSide.BUY.value, "3")
    order_3_2 = get_test_order(TradeOrderSide.BUY.value, "3", order_price=order_3[ECOBIC.PRICE.value])
    order_4 = get_test_order(TradeOrderSide.BUY.value, "4")
    order_4_2 = get_test_order(TradeOrderSide.BUY.value, "4", order_price=order_4[ECOBIC.PRICE.value])
    order_6 = get_test_order(TradeOrderSide.SELL.value, "6")
    order_6_2 = get_test_order(TradeOrderSide.SELL.value, "6", order_price=order_6[ECOBIC.PRICE.value])
    order_book_manager.handle_book_adds([
        get_test_order(TradeOrderSide.BUY.value, "1"),
        get_test_order(TradeOrderSide.BUY.value, "2"),
        order_3,
        order_4,
        get_test_order(TradeOrderSide.SELL.value, "5"),
        order_6
    ])

    assert get_order_at_id_in_order_list("1", order_book_manager.bids)
    assert get_order_at_id_in_order_list("2", order_book_manager.bids)
    assert get_order_at_id_in_order_list("3", order_book_manager.bids)
    assert get_order_at_id_in_order_list("4", order_book_manager.bids)
    assert get_order_at_id_in_order_list("5", order_book_manager.asks)
    assert get_order_at_id_in_order_list("6", order_book_manager.asks)
    # unknown order
    order_book_manager.handle_book_updates([get_test_order(TradeOrderSide.SELL.value, "10")])
    assert get_order_at_id_in_order_list("1", order_book_manager.bids)
    assert get_order_at_id_in_order_list("2", order_book_manager.bids)
    assert get_order_at_id_in_order_list("3", order_book_manager.bids)
    assert get_order_at_id_in_order_list("3", order_book_manager.bids) == order_3
    order_3_1 = deepcopy(order_3)
    order_book_manager.handle_book_updates([order_3_2])
    assert get_order_at_id_in_order_list("3", order_book_manager.bids)[ECOBIC.SIZE.value] != order_3_1[ECOBIC.SIZE.value]
    assert get_order_at_id_in_order_list("3", order_book_manager.bids)[ECOBIC.SIZE.value] == order_3_2[ECOBIC.SIZE.value]
    assert get_order_at_id_in_order_list("4", order_book_manager.bids) == order_4
    assert get_order_at_id_in_order_list("6", order_book_manager.asks) == order_6
    order_4_1 = deepcopy(order_4)
    order_6_1 = deepcopy(order_6)
    order_book_manager.handle_book_updates([order_4_2, order_6_2])
    assert get_order_at_id_in_order_list("4", order_book_manager.bids)[ECOBIC.SIZE.value] != order_4_1[ECOBIC.SIZE.value]
    assert get_order_at_id_in_order_list("4", order_book_manager.bids)[ECOBIC.SIZE.value] == order_4_2[ECOBIC.SIZE.value]
    assert get_order_at_id_in_order_list("6", order_book_manager.asks)[ECOBIC.SIZE.value] != order_6_1[ECOBIC.SIZE.value]
    assert get_order_at_id_in_order_list("6", order_book_manager.asks)[ECOBIC.SIZE.value] == order_6_2[ECOBIC.SIZE.value]


def get_test_order(order_side, order_id, order_price=None, order_size=None):
    return {
        ECOBIC.SIDE.value: order_side,
        ECOBIC.SIZE.value: order_size if order_size is not None else random_quantity(),
        ECOBIC.PRICE.value: order_price if order_price is not None else random_price(),
        ECOBIC.ORDER_ID.value: order_id
    }


def get_order_at_id_in_order_list(order_id, order_list):
    for order in order_list.values():
        if order[0][ECOBIC.ORDER_ID.value] == order_id:
            return order[0]
    return None




#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as trading_enums
import octobot_trading.constants as constants
import octobot_trading.personal_data as personal_data


_LOSING_ORDER_TYPES = [
    trading_enums.TradeOrderType.STOP_LOSS,
    trading_enums.TradeOrderType.STOP_LOSS_LIMIT,
]


def compute_win_rate(exchange_manager):
    lost_trades_count = constants.ZERO
    won_trades_count = constants.ZERO
    if exchange_manager.is_future:
        for transaction in exchange_manager.exchange_personal_data.transactions_manager.transactions.values():
            if isinstance(transaction, personal_data.RealisedPnlTransaction) and transaction.is_closed_pnl():
                if (transaction.side is trading_enums.PositionSide.SHORT and
                        transaction.average_entry_price > transaction.average_exit_price) or (
                    transaction.side is trading_enums.PositionSide.LONG and
                        transaction.average_entry_price < transaction.average_exit_price
                ):
                    won_trades_count += constants.ONE
                else:
                    lost_trades_count += constants.ONE
    else:
        for trade in exchange_manager.exchange_personal_data.trades_manager.trades.values():
            if trade.status is trading_enums.OrderStatus.FILLED and trade.side is trading_enums.TradeOrderSide.SELL:
                if trade.exchange_trade_type in _LOSING_ORDER_TYPES:
                    lost_trades_count += constants.ONE
                else:
                    won_trades_count += constants.ONE
    total_counted_trades = won_trades_count + lost_trades_count
    if total_counted_trades > constants.ZERO:
        return won_trades_count / total_counted_trades
    return constants.ZERO



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import numpy as np

from octobot_commons.enums import PriceIndexes
from octobot_trading.exchange_data.ohlcv.candles_manager import CandlesManager


def test_constructor():
    candles_manager = CandlesManager()
    assert candles_manager.candles_initialized is False
    assert candles_manager.close_candles_index == 0
    assert len(candles_manager.close_candles) == CandlesManager.MAX_CANDLES_COUNT
    assert all(np.isnan(value) for value in candles_manager.close_candles)


def test_add_new_candle():
    candles_manager = CandlesManager()
    candle = _gen_candles(1)[0]
    candles_manager.add_new_candle(candle)
    assert candles_manager.close_candles_index == 1
    assert len(candles_manager.close_candles) == CandlesManager.MAX_CANDLES_COUNT
    assert candles_manager.close_candles[0] == candle[PriceIndexes.IND_PRICE_CLOSE.value]


def test_add_old_and_new_candles():
    candles_manager = CandlesManager()

    # with one candle
    single_candle = _gen_candles(1)
    candles_manager.add_old_and_new_candles(single_candle)
    assert candles_manager.reached_max is False
    assert candles_manager.close_candles_index == 1
    assert len(candles_manager.close_candles) == CandlesManager.MAX_CANDLES_COUNT
    assert candles_manager.close_candles[0] == single_candle[0][PriceIndexes.IND_PRICE_CLOSE.value]

    # with many candles including first one
    many_candles = _gen_candles(10)
    candles_manager.add_old_and_new_candles(many_candles)
    assert candles_manager.reached_max is False
    assert candles_manager.close_candles_index == 10
    assert len(candles_manager.close_candles) == CandlesManager.MAX_CANDLES_COUNT
    assert candles_manager.close_candles[0] == many_candles[0][PriceIndexes.IND_PRICE_CLOSE.value]
    assert candles_manager.close_candles[9] == many_candles[9][PriceIndexes.IND_PRICE_CLOSE.value]


def test_replace_all_candles():
    candles_manager = CandlesManager()
    many_candles = _gen_candles(20)[10:]
    candles_manager.add_old_and_new_candles(many_candles)
    assert candles_manager.close_candles[0] == many_candles[0][PriceIndexes.IND_PRICE_CLOSE.value]
    assert candles_manager.close_candles[9] == many_candles[9][PriceIndexes.IND_PRICE_CLOSE.value]
    new_candles = _gen_candles(10)
    candles_manager.replace_all_candles(new_candles)
    assert candles_manager.close_candles[0] == new_candles[0][PriceIndexes.IND_PRICE_CLOSE.value]
    assert candles_manager.close_candles[9] == new_candles[9][PriceIndexes.IND_PRICE_CLOSE.value]


def test_get_symbol_prices():
    candles_manager = CandlesManager()
    candle = _gen_candles(1)[0]
    candles_manager.add_new_candle(candle)
    symbol_price = candles_manager.get_symbol_prices()
    assert symbol_price[PriceIndexes.IND_PRICE_CLOSE.value][-1] == candle[PriceIndexes.IND_PRICE_CLOSE.value]
    assert symbol_price[PriceIndexes.IND_PRICE_OPEN.value][-1] == candle[PriceIndexes.IND_PRICE_OPEN.value]
    assert symbol_price[PriceIndexes.IND_PRICE_HIGH.value][-1] == candle[PriceIndexes.IND_PRICE_HIGH.value]
    assert symbol_price[PriceIndexes.IND_PRICE_LOW.value][-1] == candle[PriceIndexes.IND_PRICE_LOW.value]
    assert symbol_price[PriceIndexes.IND_PRICE_VOL.value][-1] == candle[PriceIndexes.IND_PRICE_VOL.value]
    assert symbol_price[PriceIndexes.IND_PRICE_TIME.value][-1] == candle[PriceIndexes.IND_PRICE_TIME.value]

    second_candle = _gen_candles(2)[1]
    candles_manager.add_new_candle(second_candle)
    second_sym_price = candles_manager.get_symbol_prices(1)
    assert second_sym_price[PriceIndexes.IND_PRICE_CLOSE.value][-1] == second_candle[PriceIndexes.IND_PRICE_CLOSE.value]
    assert second_sym_price[PriceIndexes.IND_PRICE_OPEN.value][-1] == second_candle[PriceIndexes.IND_PRICE_OPEN.value]
    assert second_sym_price[PriceIndexes.IND_PRICE_HIGH.value][-1] == second_candle[PriceIndexes.IND_PRICE_HIGH.value]
    assert second_sym_price[PriceIndexes.IND_PRICE_LOW.value][-1] == second_candle[PriceIndexes.IND_PRICE_LOW.value]
    assert second_sym_price[PriceIndexes.IND_PRICE_VOL.value][-1] == second_candle[PriceIndexes.IND_PRICE_VOL.value]
    assert second_sym_price[PriceIndexes.IND_PRICE_TIME.value][-1] == second_candle[PriceIndexes.IND_PRICE_TIME.value]


def test_get_symbol_candles_data():
    candles_manager = CandlesManager()
    _test_data(candles_manager.get_symbol_close_candles(), 0, np.nan)
    _test_data(candles_manager.get_symbol_open_candles(), 0, np.nan)
    _test_data(candles_manager.get_symbol_high_candles(), 0, np.nan)
    _test_data(candles_manager.get_symbol_low_candles(), 0, np.nan)
    _test_data(candles_manager.get_symbol_time_candles(), 0, np.nan)
    _test_data(candles_manager.get_symbol_volume_candles(), 0, np.nan)

    new_candles = _gen_candles(2)
    candles_manager.add_old_and_new_candles(new_candles)
    _test_data(candles_manager.get_symbol_close_candles(), 2, new_candles[-1][PriceIndexes.IND_PRICE_CLOSE.value])
    _test_data(candles_manager.get_symbol_open_candles(), 2, new_candles[-1][PriceIndexes.IND_PRICE_OPEN.value])
    _test_data(candles_manager.get_symbol_high_candles(), 2, new_candles[-1][PriceIndexes.IND_PRICE_HIGH.value])
    _test_data(candles_manager.get_symbol_low_candles(), 2, new_candles[-1][PriceIndexes.IND_PRICE_LOW.value])
    _test_data(candles_manager.get_symbol_volume_candles(), 2, new_candles[-1][PriceIndexes.IND_PRICE_VOL.value])
    _test_data(candles_manager.get_symbol_time_candles(), 2, new_candles[-1][PriceIndexes.IND_PRICE_TIME.value])


def test_reach_max_candles_count():
    candles_manager = CandlesManager()
    all_candles = _gen_candles(candles_manager.MAX_CANDLES_COUNT + 3)
    max_candles = all_candles[0:candles_manager.MAX_CANDLES_COUNT]
    other_candles = all_candles[candles_manager.MAX_CANDLES_COUNT:]

    assert candles_manager.reached_max is False
    candles_manager.add_old_and_new_candles(max_candles)
    assert candles_manager.reached_max is True
    _test_data(candles_manager.get_symbol_close_candles(), candles_manager.MAX_CANDLES_COUNT,
               max_candles[-1][PriceIndexes.IND_PRICE_CLOSE.value])
    assert candles_manager.close_candles_index == candles_manager.MAX_CANDLES_COUNT - 1

    # should remove oldest (first) candles and insert new ones instead
    candles_manager.add_old_and_new_candles(other_candles)
    _test_data(candles_manager.get_symbol_close_candles(), candles_manager.MAX_CANDLES_COUNT,
               other_candles[-1][PriceIndexes.IND_PRICE_CLOSE.value])


def test_reach_max_candles_count_with_custom_candles_count():
    # negative value: use default MAX_CANDLES_COUNT
    candles_manager = CandlesManager(max_candles_count=-1)
    assert candles_manager.max_candles_count == candles_manager.MAX_CANDLES_COUNT

    candles_manager = CandlesManager(max_candles_count=CandlesManager.MAX_CANDLES_COUNT + 21)
    assert candles_manager.max_candles_count != candles_manager.MAX_CANDLES_COUNT
    assert candles_manager.MAX_CANDLES_COUNT == CandlesManager.MAX_CANDLES_COUNT
    all_candles = _gen_candles(candles_manager.max_candles_count + 3)
    max_candles = all_candles[0:candles_manager.max_candles_count]
    other_candles = all_candles[candles_manager.max_candles_count:]

    assert candles_manager.reached_max is False
    candles_manager.add_old_and_new_candles(max_candles)
    assert candles_manager.reached_max is True
    _test_data(candles_manager.get_symbol_close_candles(), candles_manager.max_candles_count,
               max_candles[-1][PriceIndexes.IND_PRICE_CLOSE.value])
    assert candles_manager.close_candles_index == candles_manager.max_candles_count - 1

    # should remove oldest (first) candles and insert new ones instead
    candles_manager.add_old_and_new_candles(other_candles)
    _test_data(candles_manager.get_symbol_close_candles(), candles_manager.max_candles_count,
               other_candles[-1][PriceIndexes.IND_PRICE_CLOSE.value])


def _test_data(candles_data, expected_len, expected_last_val):
    assert len(candles_data) == expected_len
    if expected_len > 0:
        assert candles_data[-1] == expected_last_val


def _gen_candles(size) -> list:
    return [_get_candle(seed) for seed in range(1, size + 1)]


def _get_candle(seed):
    return [seed, seed * 10, seed * 100, seed * 1000, seed * 10000, seed * 100000]
